import requests
import json
from pprint import pprint
from typing import List, Union, Any
from rara_meta_extractor.config import LOGGER

class LlamaAPIConnector:
    # TODO: parameetrid kontrollida -https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md
    def __init__(self,
            host_url: str, #e.g. "http://localhost:8080"
            stream: bool = False,
            n_predict: int = 400,
            temperature: float = 0.7,
            stop: List[str] = ["</s>","Llama:","User:"],
            repeat_last_n: int = 256,
            repeat_penalty: float = 1.18,
            penalize_nl: bool = False,
            top_k: int = 40,
            top_p: float = 0.95,
            min_p: float = 0.05,
            tfs_z: int = 1,
            typical_p: int = 1,
            presence_penalty: float = 0, #?
            frequency_penalty: float = 0, #?
            mirostat: float = 0, #?
            mirostat_tau: int = 5,
            mirostat_eta: float = 0.1,
            n_probs: float = 0, #?
            min_keep: float = 0, #?
            cache_prompt: bool = True,
            api_key: str = ""
    ):
        self.host_url = host_url
        self.stream = stream
        self.n_predict = n_predict
        self.temperature = temperature
        self.stop = stop
        self.repeat_last_n = repeat_last_n
        self.repeat_penalty = repeat_penalty
        self.penalize_nl = penalize_nl
        self.top_k = top_k
        self.top_p = top_p
        self.min_p = min_p
        self.tfs_z = tfs_z
        self.typical_p = typical_p
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.mirostat = mirostat
        self.mirostat_tau = mirostat_tau
        self.mirostat_eta = mirostat_eta
        self.n_probs = n_probs
        self.min_keep = min_keep
        self.cache_prompt = cache_prompt
        self.api_key = api_key

    @property
    def query_params(self) -> dict:
        """Convert the instance's attributes to a dictionary.
        """
        params = self.__dict__.copy()
        params.pop("host_url")
        return params

    def query(self, instructions: str, context: str, json_schema: dict = {},
                output_full_response: bool = False, default_response: Any = {}
    ) -> Any:
        """ Query information from a Llama model via an API call.

        Parameters
        -----------
        instructions: str
            Instructions for the model (system prompt).
        context: str
            User query (text).
        json_schema: dict
            Specifies output format / fieldsself.
        output_full_response: bool
            If enabled, the whole API response is added to the output;
            otherwise only "content" (the "answer").
        default_response: Any
            Default output in case Llama generates an invalid JSON string
            which cannot be loaded.

        Returns
        -----------
        Any:
            The answer to user's query (can be any type, especially when
            using a fine-tuned model or specific JSON schema) or the
            whole API response (dict).
        """
        prompt = f"{instructions}\n\nUser: {context}\nLlama:"
        payload = self.query_params
        payload["prompt"] = prompt
        if json_schema:
            payload["json_schema"] = json_schema

        url = f"{self.host_url}/completion"

        response = requests.post(url, data=json.dumps(payload))
        extracted_data = {}
        if response.status_code == 200:
            json_response = response.json()
            try:
                content = json_response.pop("content")
                content_json = json.loads(content)
            except Exception as e:
                LOGGER.error(f"Loading JSON from response failed with error: {e}. Response: '{content}'")
                content_json = default_response
            if output_full_response:
                json_response["content"] = content_json
                output = json_response
            else:
                output = content_json
        else:
            LOGGER.error(f"Status: {response.status_code}")
            LOGGER.error(response.text)
        return output



from rara_meta_extractor.tools.utils import load_txt, load_json
from rara_meta_extractor.constants.schemas import TEXT_CLASSIFIER_SCHEMA
from rara_meta_extractor.constants.data_classes import (
    TextBlock, AuthorField, TextPartLabel, TEXT_BLOCKS, META_YEAR_FIELDS
)
import importlib.resources
import os
import logging


LOGGER = logging.getLogger("rara-meta-extractor")

METADATA_TEXT_BLOCKS = [TextBlock.METADATA, TextBlock.TITLE_PAGE]

# File paths
AUTHOR_FIELDS_FILE_PATH =  importlib.resources.files("rara_meta_extractor.data.fields") / "kata_author_fields.txt"
META_FIELDS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.fields") / "kata_meta_fields.txt"

META_EXTRACTOR_INSTRUCTIONS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.instructions") / "meta_extractor_instructions.txt"
TEXT_CLASSIFIER_INSTRUCTIONS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.instructions") / "text_classifier_instructions.txt"

META_SCHEMA_PATH = importlib.resources.files("rara_meta_extractor.data.schemas") / "meta_schema.json"

AUTHOR_ROLES_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.field_translations") / "author_roles.json"

EPUB_AUTHOR_ROLES_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.field_translations") / "relator_to_est.json"
EPUB_META_KEYS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.fields") / "epub_meta_keys.txt"

AUTHOR_ROLES_DICT = load_json(AUTHOR_ROLES_FILE_PATH)

EPUB_AUTHOR_ROLES_DICT = load_json(EPUB_AUTHOR_ROLES_FILE_PATH)

AUTHOR_FIELDS = load_txt(AUTHOR_FIELDS_FILE_PATH, to_list=True)
META_FIELDS = load_txt(META_FIELDS_FILE_PATH, to_list=True)

EPUB_META_KEYS = load_txt(EPUB_META_KEYS_FILE_PATH, to_list=True)


LLAMA_META_EXTRACTOR_INSTRUCTIONS = load_txt(META_EXTRACTOR_INSTRUCTIONS_FILE_PATH)
LLAMA_TEXT_CLASSIFIER_INSTRUCTIONS = load_txt(TEXT_CLASSIFIER_INSTRUCTIONS_FILE_PATH)

META_SCHEMA = load_json(META_SCHEMA_PATH)


LLAMA_TEXT_CLASSIFIER_URL = "http://dev-elastic1.texta.ee:8080"
LLAMA_META_EXTRACTOR_URL = "http://localhost:8080"

FIELDS = META_FIELDS + AUTHOR_FIELDS


DEFAULT_INSTRUCTIONS = "This is a conversation between User and Llama - a metadata extrator with skills of professional cataloguer. Please extract the following fields: {0}"

KATA_INSTRUCTIONS = (
    f"This is a conversation between User and Llama - a metadata extrator with skills of professional cataloguer. " +
    f"Your only assignment is to extract relevant metadata and OUTPUT IT AS JSON DICT. " +
    f"THIS IS VERY IMPORTANT! Output it as JSON DICT or DIE. Fields you are trying to extract are the following: {FIELDS}\n"
)

META_EXTRACTOR_CONFIG = {
    "llama_host_url": LLAMA_META_EXTRACTOR_URL,
    "instructions": KATA_INSTRUCTIONS,
    "fields": [],
    "json_schema": {},
    "temperature": 0.0,
    "n_predict": 500
}

TEXT_CLASSIFIER_CONFIG = {
    "llama_host_url": LLAMA_TEXT_CLASSIFIER_URL,
    "instructions": LLAMA_TEXT_CLASSIFIER_INSTRUCTIONS,
    "fields": TEXT_CLASSIFIER_SCHEMA,
    "json_schema": {},
    "temperature": 0.0,
    "n_predict": 500
}



from typing import List, Any
from pprint import pprint
from rara_meta_extractor.config import LOGGER, DEFAULT_INSTRUCTIONS
from rara_meta_extractor.llama_api_wrapper import LlamaAPIConnector
from rara_meta_extractor.tools.llama_schema_generator import JSONSchemaGenerator



class LlamaExtractor:
    def __init__(
        self,
        llama_host_url: str,
        instructions: str = DEFAULT_INSTRUCTIONS,
        temperature: float = 0.7,
        n_predict: int = 500,
        json_schema: dict = {},
        fields: List[str] | List[dict] = [],
        **kwargs
    ):
        self.llama_host_url: str = llama_host_url
        self.fields_schema: List[str] | List[dict] = fields

        self.llama_connector: LlamaAPIConnector = LlamaAPIConnector(
            host_url=llama_host_url,
            temperature=temperature,
            n_predict=n_predict,
            **kwargs
        )

        self.schema_generator: JSONSchemaGenerator = JSONSchemaGenerator()
        self.__fields: List[str]  = []
        self.__fields_str: str = ""
        self.__instructions: str = instructions.format(self.fields_str)
        self.__json_schema: dict = json_schema


    @property
    def fields(self) -> List[str]:
        if not self.__fields:
            if self.fields_schema and isinstance(self.fields_schema[0], dict):
                self.__fields = [
                    field.get("name")
                    for field in self.fields_schema
                ]
            else:
                self.__fields = self.fields_schema
        return self.__fields

    @property
    def fields_str(self) -> str:
        if not self.__fields_str:
            self.__fields_str = ", ".join(self.fields)
        return self.__fields_str

    @property
    def instructions(self) -> str:
        return self.__instructions

    @property
    def json_schema(self) -> str:
        if not self.__json_schema:
            if self.fields_schema:
                self.__json_schema = self.schema_generator.generate_json_schema(
                    fields=self.fields_schema
                )
        return self.__json_schema

    def extract(self, text: str, instructions: str = "",
            output_full_response: bool = False, default_response: Any = {}
    ) -> dict:
        """ Extracts information from the input text.

        Parameters
        -----------
        text: str
            Text from where to extract information.
        instructions: str
            Instructions for the LLM.
        output_full_response: bool
            If enabled, the whole API response is added to the output;
            otherwise only "content" (the "answer").
        default_response: Any
            Default output in case Llama generates an invalid JSON string
            which cannot be loaded.

        Returns
        -----------
        Any:
            LLM API resonse (or content). Usually a JSON dict.
        """
        if not instructions:
            instructions = self.instructions

        extracted_data = self.llama_connector.query(
            instructions=instructions,
            context=text,
            json_schema=self.json_schema,
            output_full_response=output_full_response,
            default_response=default_response
        )
        return extracted_data



import os
import json
from pprint import pprint
from typing import Dict, List
from bs4 import BeautifulSoup, Tag
from copy import deepcopy
from rara_meta_extractor.tools.meta_formatter import Author, Title, Meta
# TODO: rename EPUB_AUTHOR_ROLES_DICT
from rara_meta_extractor.config import (
    EPUB_AUTHOR_ROLES_DICT, EPUB_META_KEYS, LOGGER
)
from rara_meta_extractor.constants.data_classes import (
    MetaField, AuthorField, TitleTypeSimple, TitleType, ISSUE_STYLE_MAP
)


class MetsAltoMetaExtractor:
    def __init__(self):
        self.relator_to_est: dict = EPUB_AUTHOR_ROLES_DICT

        # Needed subitems from METS ALTO meta fields
        #(don't want separate metadata for pictures, chapters/articles)
        self.relevant_subitems = ["PRINT", "ELEC", "ISSUE"]

    def _get_role_equivalent(self, role_candidate: str) -> str:
        """ Find MARC relator roles equivalents, used only for parsing
        authors/contributors from sections_meta fields.
        If role is not in dictionary, then role is unknown
        """
        role = self.relator_to_est.get(role_candidate, AuthorField.UNKNOWN)
        return role


    def _parse_section_meta(self, sec_meta: Tag|str, seq_nr: str,
        simple: bool = False
    ) -> dict:
        """ Parse section meta.
        """
        section_parsed = {"sequence_number": seq_nr}

        if isinstance(sec_meta, str):
            sec_meta = BeautifulSoup(sec_meta, "lxml-xml")

        if sec_meta.title:
            title = [tag.string for tag in sec_meta.find_all("title")]
            section_parsed["title"] = title

        if sec_meta.titleInfo:
            if not isinstance(sec_meta.titleInfo, str):
                if "xml:lang" in sec_meta.titleInfo.attrs.keys():
                    title_language = sec_meta.titleInfo["xml:lang"]
                    section_parsed["title_language"] = title_language

        titles = section_parsed.pop("title", [])
        title_lang = section_parsed.pop("title_language", "")
        title_type = TitleType.METS_TITLE if not simple else TitleTypeSimple.METS_TITLE

        new_titles = Title(titles=titles, lang=title_lang, title_type=title_type).to_dicts(
            simple=simple
        )
        section_parsed[MetaField.TITLES] = new_titles
        kw = [tag for tag in sec_meta.find_all("subject")]

        if kw != []:
            keywords = [
                {
                    "original_keyword_value": k.topic.string
                }
                for k in kw
            ]
            section_parsed["other_keywords"] = keywords

        names = [tag for tag in sec_meta.find_all("name")]

        if names:
            authors = []
            for name in names:
                author_type = name.get("type", "")
                author_name = " ".join(
                    [
                        namePart.string
                        for namePart in name.find_all("namePart")
                    ]
                )
                author_role = self._get_role_equivalent(
                    role_candidate=name.roleTerm.string
                )
                new_author = Author(
                    name=author_name,
                    role=author_role,
                    is_primary=True,
                    author_type=author_type,
                    map_role=False
                ).to_dict(simple=simple)
                authors.append(new_author)


            # We don't know if author is primary or not, so this is
            # currently left out of the authors list,
            # however we can get the author type
            section_parsed[MetaField.AUTHORS] = authors

        if sec_meta.languageTerm:
            language_term = sec_meta.languageTerm.string
            section_parsed["language"] = language_term
        return section_parsed

    def _run_sections_parsing(self, digitized_texts: List[dict], simple: bool = False) -> List[dict]:
        """ Optional, useful for getting tables of
        content/parts of text.
        """
        sections_meta = []

        section_meta = [
            {
                "sequence_number": section["sequence_nr"],
                "section_meta": section["section_meta"]
            }
            for section in digitized_texts
            if section.get("section_meta", None)
        ]
        if section_meta != [{}]:
            for sec_meta in section_meta:
                section_meta_p = self._parse_section_meta(
                    sec_meta=sec_meta.get("section_meta"),
                    seq_nr=sec_meta.get("sequence_number"),
                    simple=simple
                )
                sections_meta.append(section_meta_p)
        return sections_meta

    def _check_if_in_dict(self, target_dict: dict, add_key: str, val: str) -> dict:
        """ Checking if value, key already present in target dictionary.
        Adding key to dictionary, if not present, appending value
        to dictionary key if not present etc.
        """
        if add_key in target_dict:
            if val not in target_dict[add_key]:
                target_dict[add_key].append(val)
        else:
            target_dict[add_key] = [val]
        return target_dict

    def _parse_issn(self, issn: str) -> str:
        parts = issn.split("-")
        parsed_issn = "".join(parts)
        return [parsed_issn]

    def extract_meta(self,
            mets_alto_metadata: List[str], digitized_texts: List[dict], simple: bool = False
    ) -> dict:

        mini_parsed_xml = {
            MetaField.TITLES: {},
            "task_task": {}
        }

        for subitem in mets_alto_metadata:

            si_soup = BeautifulSoup(subitem, "lxml-xml")
            relevant_subitems = [
                n_subitem
                for n_subitem in self.relevant_subitems
                if n_subitem in si_soup.dmdSec["ID"]
            ]

            if relevant_subitems:
                if si_soup.title != None:
                    title = si_soup.title.string
                    if title:
                        mini_parsed_xml[MetaField.TITLES] = self._check_if_in_dict(
                            target_dict=mini_parsed_xml[MetaField.TITLES],
                            add_key="title",
                            val=title
                        )

                if si_soup.partNumber:
                    part_nr = si_soup.partNumber.string
                    mini_parsed_xml[MetaField.TITLES]["part_number"] = part_nr

                if si_soup.dateIssued:
                    dateissued = si_soup.dateIssued.string
                    dateissued = str(dateissued)
                    mini_parsed_xml = self._check_if_in_dict(
                        target_dict=mini_parsed_xml,
                        add_key="publication_date",
                        val=dateissued
                    )

                if si_soup.copyrightDate:
                    datecpr = si_soup.copyrightDate.string
                    mini_parsed_xml["copyright_year"] = datecpr

                if si_soup.identifier:
                    identifier = [
                        {
                            "id": idf.string,
                            "type": idf["type"]
                        }
                        if "type" in idf.attrs.keys()
                        else {"id": idf.string, "type": "unknown"}
                        for idf in si_soup.find_all("identifier")
                    ]
                    other_ids = []
                    for id_ex in identifier:
                        if id_ex["type"] == "CatalogueIdentifier":
                            mini_parsed_xml["ester_ID"] = id_ex["id"]
                        elif id_ex["type"] == "issn":
                            mini_parsed_xml["issn"] = self._parse_issn(id_ex["id"])
                        elif id_ex["type"] == "local":
                            mini_parsed_xml["task_task"]["article_date"] = id_ex["id"]
                        else:
                            other_ids.append(id_ex["id"])

                    if other_ids != []:
                        mini_parsed_xml["ID"] = other_ids

                if si_soup.typeOfResource:
                    resource_type = si_soup.typeOfResource.string
                    if "content_type" not in mini_parsed_xml:
                        mini_parsed_xml["content_type"] = resource_type

                if si_soup.titleInfo:
                    if not isinstance(si_soup.titleInfo, str):
                        if "xml:lang" in si_soup.titleInfo.attrs.keys():
                            title_language = si_soup.titleInfo["xml:lang"]
                            if "title_language" not in mini_parsed_xml[MetaField.TITLES]:
                                mini_parsed_xml[MetaField.TITLES]["title_language"] = title_language

                if si_soup.genre:
                    genre = [
                        genre.string
                        for genre in si_soup.find_all("genre")
                        if genre.string!="\n"
                    ]
                    if genre:
                        for g in genre:
                            mini_parsed_xml["task_task"] = self._check_if_in_dict(
                                target_dict=mini_parsed_xml["task_task"],
                                add_key="style",
                                val=g
                            )
        sections_meta_parsed = self._run_sections_parsing(digitized_texts, simple=simple)
        mini_parsed_xml.update(
            {
                "sections": sections_meta_parsed
            }
        )

        titles = mini_parsed_xml.pop(MetaField.TITLES, {})
        new_titles = Title(
            titles=titles.get("title", []),
            lang=titles.get("title_language", ""),
            part_number=titles.get("part_number", "")
        ).to_dicts(simple=simple)
        mini_parsed_xml[MetaField.TITLES] = new_titles
        task_task = mini_parsed_xml.pop("task_task")
        raw_issue_types = task_task.get("style", [])
        if raw_issue_types:
            issue_type = ISSUE_STYLE_MAP.get(raw_issue_types[0], "")
        mini_parsed_xml[MetaField.ISSUE_TYPE] = issue_type
        mini_parsed_xml = Meta.update_field_types(
            meta_dict=mini_parsed_xml,
            custom_keys_to_ignore=["sections"]
        )

        return mini_parsed_xml



from typing import Dict, Any, List, Tuple
from pprint import pprint
import regex as re
import json
import typing
from copy import deepcopy
from rara_meta_extractor.config import (
    EPUB_AUTHOR_ROLES_DICT, EPUB_META_KEYS, LOGGER
)
from rara_meta_extractor.constants.data_classes import MetaField
from rara_meta_extractor.tools.meta_formatter import Author, Title, Meta

class EPUBMetaExtractor:
    def __init__(self):
        self.relator_to_est: dict = EPUB_AUTHOR_ROLES_DICT
        self.meta_keys: List[str] = EPUB_META_KEYS
        self.metadata_set = {
            meta_key: {}
            for meta_key in self.meta_keys
        }
        self.find_script = re.compile(r"<[^>]*>")
        self.cover_name = re.compile(r"[^.]*.(xht|ht|x)ml")
        self.isbn_description = re.compile(
            r"ISBN ^(?=(?:\D*\d){10}(?:(?:\D*\d){3})?$)[\d-]+$"
        )


    def _get_values(self, data_list: list) -> List[Any]:
        values = [
            elem.get("value", elem)
            for elem in data_list
        ]
        return values

    def _replace_key(self, url_key: str) -> str:
        if "}" in url_key:
            new_key = url_key.split("}")[-1]
            return new_key
        else:
            return url_key

    def _iterate_until_find_data_dict(self, structure: dict) -> dict:
        structure_data = {}
        for structure_item in structure:
            if isinstance(structure_item, str):
                structure_data["value"] = structure_item
            else:
                for tuple_item in structure_item:
                    if isinstance(tuple_item, str):
                        structure_data["value"] = tuple_item
                    elif isinstance(tuple_item, dict):
                        if len(tuple_item) > 0:
                            [
                                structure_data.update(
                                    {self._replace_key(key): value}
                                )
                                for key, value in tuple_item.items()
                            ]
        return structure_data

    def _iterate_until_find_data_list(self,
            structure: List[Any]
    ) -> List[dict]:
        structure_data = []
        for structure_item in structure:
            if None not in structure_item:
                temp_data = {}
                if isinstance(structure_item, str):
                    if structure_item is not None:
                        temp_data["value"] = structure_item
                else:
                    for tuple_item in structure_item:
                        if isinstance(tuple_item, str):
                            temp_data["value"] = tuple_item
                        elif isinstance(tuple_item, dict):
                            if len(tuple_item) > 0:
                                [
                                    temp_data.update(
                                        {self._replace_key(key): value}
                                    )
                                    for key, value in tuple_item.items()
                                ]
                structure_data.append(temp_data)
        return structure_data

    def _iterate_until_find_data(self, structure: List[Any] | dict ) -> list:
        # Seeing if multiple values inside the structure
        if len(structure) == 1:
            structure_data = self._iterate_until_find_data_dict(structure)
            return [structure_data]
        else:
            structured_data_list = []
            temp_dict = self._iterate_until_find_data_list(structure)
            structured_data_list.extend(temp_dict)
            return structured_data_list


    def _parse_authors(self,
            authors: List[dict], is_primary: bool,
            unk_key: str = "unknown", simple: bool = False
    ) -> List[dict]:
        parsed_authors = []
        for author in authors:
            original_role = author.get("role", "")
            role = self.relator_to_est.get(original_role, unk_key)
            name = author.get("value", author)

            author_structure = Author(
                name=name,
                is_primary=is_primary,
                role=role,
                map_role=False
            ).to_dict(simple=simple)

            if role == unk_key:
                LOGGER.debug(f"Could not detect role for author: {author}.")

            parsed_authors.append(author_structure)
        return parsed_authors

    def _authors_correct_structure(self,
            parsed_metadata: dict, unk_key: str = "unknown", simple: bool = False
    ) -> List[dict]:
        authors = []

        creators = parsed_metadata.get("creator", [])
        contributors = parsed_metadata.get("contributor", [])

        parsed_creators = self._parse_authors(
            authors=creators, is_primary=True, simple=simple
        )
        parsed_contributors = self._parse_authors(
            authors=contributors, is_primary=False, simple=simple
        )

        authors = parsed_creators + parsed_contributors
        return authors

    def _get_publication_time(self, dates: List[dict]) -> List[str] | str:
        publication_time = [
            date.get("value")
            for date in dates
        ]
        if len(publication_time) == 1:
            publication_time = publication_time[0]
        return publication_time

    def _process_date(self, parsed_metadata: dict) -> dict:
        relevant_dates = ["creation", "publication"]
        date_info = {}
        dates = parsed_metadata.get("date", [])
        if len(dates) == 1:
            publication_time = self._get_publication_time(
                dates=dates
            )

        else:
            events = [
                date for date in dates
                if "event" in date
            ]
            if events:
                publication_creation = [
                    date for date in events
                    if date.get("event", "") in relevant_dates
                ]
                if publication_creation:
                    publication_time = self._get_publication_time(
                        dates=publication_creation
                    )

                else:
                    publication_time = self._get_publication_time(
                        dates=events
                    )
            else:
                publication_time = self._get_publication_time(
                    dates=dates
                )

        date_info["publication_date"]  = publication_time # TODO!
        return date_info

    def _identify_date(self,
            parsed_metadata: dict, unk_key: str = "unknown"
    ) -> dict:
        date_info = {"publication_date": unk_key}

        dates = parsed_metadata.get("date", [])
        timestamps = parsed_metadata.get("timestamp", [])

        if dates:
            parsed_date = self._process_date(parsed_metadata)
            date_info.update(parsed_date)
            publication_time = date_info.get("publication_date")

            if "0101-01-01" in publication_time and timestamps:
                timestamp = timestamps[0].get("content", "")
                if timestamp and timestamp != "0101-01-01T00:00:00+00:00":
                    date_info["publication_date"] = timestamp

        elif timestamps:
            timestamp = timestamps[0].get("content", "")
            if timestamp and timestamp != "0101-01-01T00:00:00+00:00":
                date_info["publication_date"] = timestamp # TODO!!!!
        return date_info

    def get_series_data(self, parsed_metadata: dict) -> dict:
        series_info = {}
        series = parsed_metadata.get("series", [])
        if series:
            # TODO!!!!
            series_info["added_series_volume"] = str(int(float(
                parsed_metadata["series_index"][0]["content"]
            )))
            series_info["added_series_title"] = series[0]["content"]
        return series

    def _correct_structure_other_metadata(self,
            parsed_metadata: dict, other_meta: dict
    ) -> dict:
        other_vals = {}
        if parsed_metadata["identifier"] != {}:
            for identifier in parsed_metadata["identifier"]:
                if "value" in identifier:
                    if "id" in identifier:
                        if identifier["id"] in ["ISBN", "isbn-id"]:
                            other_vals["isbn"] = [identifier["value"]]

                        if identifier["id"] == "ISSN":
                            other_vals["issn"] = [identifier["value"]]

                        if identifier["id"].lower() in ["bookid", "book-id", "uuid_id"]:
                            if "ID" not in other_vals:
                                other_vals["ID"] = [identifier["value"]]
                            else:
                                other_vals["ID"] = other_vals["ID"].append(
                                    identifier["value"]
                                )
                    elif "scheme" in identifier:
                        if identifier["scheme"] == "ISBN":
                            other_vals["isbn"] = [identifier["value"]]

                        if identifier["scheme"] == "ISSN":
                            other_vals["issn"] = [identifier["value"]]

                        if identifier["scheme"].lower() in ["bookid", "uuid", "uid"]:
                            if "ID" not in other_vals:
                                other_vals["ID"] = [identifier["value"]]
                            else:
                                other_vals["ID"] = other_vals["ID"].append(
                                    identifier["value"]
                                )
        if parsed_metadata["source"] != {}:
            for source in parsed_metadata["source"]:
                if "id" in source:
                    if source["id"] == "src-id":
                        if "ID" not in other_vals:
                            other_vals["ID"] = [source["value"]]
                        else:
                            other_vals["ID"] = other_vals["ID"].append(
                                source["value"]
                            )
        series = self.get_series_data(parsed_metadata)
        if series != {}:
            other_vals["series"] = series

        if parsed_metadata["publisher"] != {}:
            other_vals["publisher"] = [
                item["value"] if "value" in item else item
                for item in parsed_metadata["publisher"]
            ]
        if parsed_metadata["format"] != {}:
            other_vals["task_task"] = {
                "file_type": parsed_metadata["format"][0]["value"]
            }
        if len(parsed_metadata["subject"]) > 0:
            if len(parsed_metadata["subject"][0]) > 0:
                other_vals["uncontrolled_index_termin"] = [
                    item["value"]
                    for item in parsed_metadata["subject"]
                ]
        if len(parsed_metadata["type"]) > 0:
            if len(parsed_metadata["type"][0]) > 0:
                other_vals["uncontrolled_index_termin"] = [
                    item["value"]
                    for item in parsed_metadata["type"]
                ]
        if len(parsed_metadata["rights"]) > 0:
            other_vals["copyright_year"] = [
                item["value"]
                if "value" in parsed_metadata["rights"] else item
                for item in parsed_metadata["rights"]
            ]
        if other_meta != {}:
            [
                other_vals[key].append(val)
                for key, val in other_meta.items()
                if key in other_vals and val not in other_vals[key]
            ]
        return other_vals

    def _clean_script_description(self, description: str) -> str:
        if self.find_script.search(description):
            subbed_description = self.find_script.sub("", description)
            return subbed_description
        else:
            return description

    def _parse_isbn_description(self, description: str) -> str | None:
        if self.isbn_description.match(description):
            return None
        else:
            return description

    def _clean_descriptions(self, included_text: str|list, correct_structure: dict) -> dict:
        if isinstance(included_text, list):
            descriptions = []
            for description in included_text:
                if description != {}:
                    description = description.strip()
                    if len(description) > 3:
                        # Description contains
                        if "<" in description:
                            descriptions.append(
                                self._clean_script_description(description)
                            )
                        # Description refers to cover file
                        elif self.cover_name.match(description):
                            pass
                        elif description.startswith("ISBN"):
                            if "ISBN" not in correct_structure:
                                isbn = self._parse_isbn_description(description)
                                if isbn:
                                    correct_structure["ISBN"] = description.split()[-1]
                                else:
                                    descriptions.append(description)
                        else:
                            descriptions.append(description)
            if descriptions != []:
                correct_structure["task_includedtext"] = {
                    "text_value": descriptions
                }

        elif isinstance(included_text, str):
            description = included_text.strip()
            if len(description) > 3:
                # Description likely contains script
                if "<" in description:
                    correct_structure["task_includedtext"] = {
                        "text_value": self._clean_script_description(description)
                    }
                # Description refers to cover file
                elif self.cover_name.match(description):
                    pass
                elif description.startswith("ISBN"):
                    if "ISBN" not in correct_structure:
                        isbn = self._parse_isbn_description(description)
                        if isbn:
                            correct_structure["ISBN"] = description.split()[-1]
                        else:
                            correct_structure["task_includedtext"] = {
                                "text_value": description
                            }
                else:
                    correct_structure["task_includedtext"] = {
                        "text_value": description
                    }
        return correct_structure

    def postfilter_authors(self, authors: list) -> Tuple[Dict, list]:
        other_meta = {}
        new_authors = []
        for author in authors:
            if author["role"] in ["väljaandja", "tootja", "levitaja"]:
                if author["role"] == "väljaandja":
                    other_meta[MetaField.MANUFACTURER] = author["name"]
                elif author["role"] == "tootja":
                    other_meta[Metafield.PUBLISHER] = author["name"]
                elif author["role"] == "levitaja":
                    other_meta[MetaField.DISTRIBUTER_NAME] = author["name"]
            else:
                new_authors.append(author)
        return new_authors, other_meta

    def get_included_text(self, parsed_metadata: dict)-> list | str:
        descriptions_info = []

        descriptions = parsed_metadata.get("description", [])
        coverages = parsed_metadata.get("coverage", [])

        parsed_descriptions = self._get_values(descriptions)
        parsed_coverages = self._get_values(coverages)

        descriptions_info = parsed_descriptions + parsed_coverages
        if len(descriptions_info) == 1:
            included_texts = descriptions_info[0]
        else:
            included_texts = descriptions_info
        return included_texts

    def get_titles(self, parsed_metadata: dict) -> dict:
        task_title = {}
        titles = [
            item.get("value", item)
            for item in parsed_metadata.get("title", [])
        ]
        title_sort = parsed_metadata.get("title_sort", [])
        if len(title_sort) > 0:
            for title in title_sort:
                value = title.get("content")
                if value not in titles:
                    titles.append(value)
        task_title["title"] = titles # TODO: move key somewhere else?
        return task_title


    def correct_structure(self, parsed_metadata: dict, simple: bool = False) -> dict:
        # Language should be coming from digitizer,
        # ignore ebook language for now
        correct_structure = {
            "task_title": self.get_titles(parsed_metadata)
        }
        authors, other_meta = self.postfilter_authors(
            authors=self._authors_correct_structure(parsed_metadata, simple=simple)
        )
        correct_structure.update(
            {"task_author": authors}
        )
        correct_structure.update(
            self._correct_structure_other_metadata(
                parsed_metadata=parsed_metadata,
                other_meta=other_meta
            )
        )
        correct_structure.update(
            self._identify_date(parsed_metadata)
        )
        included_text = self.get_included_text(parsed_metadata)
        if included_text != [] and included_text != {}:
            correct_structure = self._clean_descriptions(
                included_text=included_text,
                correct_structure=correct_structure
            )
        return correct_structure

    def parse_ebook_metadata(self, metadata: dict) -> dict:
        metadata_set = deepcopy(self.metadata_set)
        for metadata_item in metadata_set:
            for value in metadata:
                if metadata_item in value:
                    metadata_set[metadata_item] = self._iterate_until_find_data(
                        structure=value[metadata_item]
                    )
        return metadata_set

    def _reformat(self, item_correct_structure: dict, simple: bool = False) -> dict:
        # Hotfix
        key_map = {
            "task_author": MetaField.AUTHORS,
            "task_title": MetaField.TITLES
        }
        new_structure = {}
        for key, value in list(item_correct_structure.items()):
            key = key_map.get(key, key)
            if key == MetaField.TITLES:
                value = Title(value.get("title", [])).to_dicts(simple=simple)
            new_structure[key] = value
        new_structure = Meta.update_field_types(new_structure)
        return new_structure


    def extract_meta(self, epub_metadata: dict, simple: bool = False) -> dict:
        # is the type correct????
        metadata_from_digitizer = list(epub_metadata.values())
        item_parsed = self.parse_ebook_metadata(metadata_from_digitizer)
        item_correct_structure = self.correct_structure(item_parsed, simple=simple)
        # hotfix:
        fields_to_ignore = ["ID", "copyright_year", "task_includedtext"]
        for field in fields_to_ignore:
            item_correct_structure.pop(field, "")
        item_correct_structure = self._reformat(item_correct_structure, simple=simple)

        return item_correct_structure



import os
import json
import jsonlines
import langdetect
from typing import List, Optional

def load_txt(file_path: str, to_list: Optional[bool] = False) -> List[str] | str:
    """ Loads text from a given file_path.

    Parameters
    -----------
    file_path: str
        Path to file containing the text.
    to_list: Optional[bool]
        If enabled, each row is treated as a list element
        and a list of strings is returned.

    Returns
    -----------
    out: List[str] | str

    """
    with open(file_path, "r") as f:
        out = f.read().strip()
        if to_list:
            out = [
                row.strip()
                for row in out.split("\n")
                if row.strip()
            ]
    return out

def load_json(file_path: str) -> list | dict:
    """ Loads a JSON file.
    """
    with open(file_path, "r") as f:
        data = json.load(f)
    return data

def dump_json(data, file_path: str):
    """ Dump data into a JSON file.
    """
    with open(file_path, "w") as f:
        json.dump(data, f)

def jl_generator(file_path: str):
    """ Iterates over a JSON-Lines file and
    yields rows.
    """
    with jsonlines.open(file_path, "r") as f:
        for row in f:
            yield row

def write_line(doc: dict, jl_file_path: str):
    """ Appends documents into a JSON-Lines file.
    """
    with jsonlines.open(jl_file_path, "a") as f:
        f.write(doc)

def detect_language(text: str) -> str:
    """ Detects language of the text.
    """
    try:
        lang = langdetect.detect(text)
    except:
        lang = ""
    return lang



import json
from typing import List, Tuple, Dict
from rara_meta_extractor.config import LOGGER
from rara_meta_extractor.constants.data_classes import TextPartLabel
from rara_meta_extractor.text_part_classifiers.base_text_part_classifier import BaseTextPartClassifier

class EPUBTextPartClassifier(BaseTextPartClassifier):
    def __init__(self):
        super().__init__()
        self.used_sections: list = []
        self.section_title_filter: List[str] = [
            "peatükk", "stseen", "vaatus", "osa", "jagu",
            "üks", "kaks", "kolm", "esimene", "teine", "kolmas"
        ]
        self.section_title_filter_roman: List[str] = [
            "II", "IV", "IX", "ii", "iv",
            "ix", "Ii", "Iv", "Ix", "XII"
        ]
        self.section_title_filter_numbers: List[str] = [
            "1", "2", "3", "4", "5"
        ]
        self.conclusion_upper_keywords: list = []

    def _toc_from_section_titles(self, digitized_texts: List[dict])-> List[dict]:
        """ Trying to put together a table of contents
        for ebooks through section titles.
        """
        # Filter for getting rid of nondescript chapter names
        # we don't want in out table of contents
        toc = []
        has_title = [
            txt for txt in digitized_texts
            if "section_title" in txt
        ]

        title_not_none = [
            txt for txt in has_title
            if txt["section_title"] != None
        ]

        if len(set([section["section_title"] for section in title_not_none])) > 1:
            for section in title_not_none:
                stripped_title = section["section_title"].strip().strip(",.")
                if stripped_title in self.section_title_filter_roman:
                    return toc
                if stripped_title in self.section_title_filter_numbers:
                    return toc
                for filter_item in self.section_title_filter:
                    if filter_item in stripped_title.lower():
                        return toc
            for section in title_not_none:
                if section["section_title"] not in toc:
                    toc.append(section["section_title"])
        return toc

    def _ebooks_get_conclusion_from_meta(self, parsed_meta: dict )-> list:
        """ Using different methods to find conclusions.
        """
        if "task_includedtext" in parsed_meta:
            if parsed_meta["task_includedtext"] != None:
                return [parsed_meta["task_includedtext"]]
        return []

    def get_conclusions(self, digitized_texts: List[dict], epub_meta: dict = {}):
        """ Using different methods to find conclusions.
        """
        LOGGER.debug(f"Searching for conclusions.")
        concl = self._look_through_section_titles(
            digitized_texts=digitized_texts,
            search_type=TextPartLabel.CONCLUSION
        )
        if not concl:
            concl = self._look_through_texts(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.CONCLUSION
            )

        # Second option: get conclusion info
        # from parsed_meta (if exists)
        if not concl:
            # ... or whatever key parsed meta is saved as
            if epub_meta:
                # No language data in parsed metadata, so used as second option
                concl = self._ebooks_get_conclusion_from_meta(
                    parsed_meta=epub_meta
                )
        return concl

    def get_tables_of_content(self, digitized_texts: List[dict]):
        """ Using different methods to find tables of content
        for ebook files.
        """
        LOGGER.debug(f"Searching for tables_of_content.")
        # Since ebooks don't have authors names associated with chapters,
        # we first try to fetch section titles by themselves
        toc = self._toc_from_section_titles(digitized_texts=digitized_texts)
        if not toc:
            # After this we look at where the toc should be
            # according to section titles and extract the text
            toc = self._look_through_section_titles(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.TABLE_OF_CONTENTS
            )
        if not toc:
            # After this we look at the text itself using keywords
            toc = self._look_through_texts(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.TABLE_OF_CONTENTS
            )
        return toc

    def get_abstracts(self, digitized_texts: List[dict]):
        """ Using different methods to find abstracts.
        """
        LOGGER.debug(f"Searching for abstracts.")
        abstracts = self._look_through_section_titles(
            digitized_texts=digitized_texts,
            search_type=TextPartLabel.ABSTRACT
        )
        if not abstracts:
            abstracts = self._look_through_texts(
            digitized_texts=digitized_texts,
            search_type=TextPartLabel.ABSTRACT
        )
        return abstracts

    def get_parts_of_text(self, digitized_texts: List[dict], epub_meta: dict) -> dict:
        """ Getting results for ebooks.

        Parameters
        -----------
        digitized_texts: List[dict]
            `texts` from Digitizer output.
        epub_meta: dict
            Output of `EPUBMetaExtractor.extract_meta()`.

        Returns
        -----------
        dict
            TODO
        """
        LOGGER.info(f"Classifying EPUB text sections.")
        tables_of_content = self.get_tables_of_content(
            digitized_texts=digitized_texts
        )
        conclusions = self.get_conclusions(
            digitized_texts=digitized_texts,
            epub_meta=epub_meta
        )
        abstracts = self.get_abstracts(
            digitized_texts=digitized_texts
        )
        results = self._results_correct_fields(
            {
                "table_of_contents": tables_of_content,
                "conclusions": conclusions,
                "abstracts": abstracts
            }
        )
        return results



import json
from typing import List, Tuple
from rara_meta_extractor.constants.data_classes import TextPartLabel
from rara_meta_extractor.config import LOGGER
from rara_meta_extractor.tools.meta_formatter import TextPart

class BaseTextPartClassifier:
    def __init__(self):
        self.search_types_keywords: dict = {
            TextPartLabel.TABLE_OF_CONTENTS: [
                "sisukord", "table of contents", "contents"
            ],
            TextPartLabel.CONCLUSION: [
                "kokkuvõte", "sisukokkuvõte", "conclusion"
            ],
            TextPartLabel.ABSTRACT: [
                "abstract", "abstracts", "annotatsioon"
            ]
        }

    def _make_concl_upper_keywords(self):
        """ Generating conclusion keywords with upper/title-case values
        for determining conclusions from inside section text.
        """
        upper_kw = [
            kw.upper()
            for kw in self._search_types_keywords[TextPartLabel.CONCLUSION]
        ]
        upper_kw.extend(
            [
                kw.title()
                for kw in self._search_types_keywords[TextPartLabel.CONCLUSION]
            ]
        )
        self.conclusion_upper_keywords = upper_kw


    def _remove_double_toc(self, toc_text: str) -> str:
        """ Helper functions to get rid of double tables of
        content/postprocess conclusions
        """
        ttl = toc_text.lower()
        for kw in self.search_types_keywords[TextPartLabel.TABLE_OF_CONTENTS]:
            if kw in ttl:
                if ttl.count(kw)>1:
                    double_kw = "\n"+kw
                    toc_text = toc_text[0:len(ttl.split(double_kw)[0])]
                    return toc_text
        return toc_text

    def _is_consecutive(self, list_of_ints: List[int]) -> bool:
        """ Check if conclusions are consecutive for joining.
        """
        sorted_list = sorted(list_of_ints)
        for i in range(1, len(sorted_list)):
            if sorted_list[i] != sorted_list[i-1] + 1:
                return False
        return True

    def _join_conclusions(self, consecutive_conclusions: List[dict]) -> List[dict]:
        """ Join consecutive conclusions.
        """
        language = consecutive_conclusions[0]["task_language"]
        consecutive_conclusions.sort(key=lambda d: d["sequence_nr"])
        concl_text = " \n".join(
            [
                item["text_value"]
                for item in consecutive_conclusions
            ]
        )
        concl = [
            {
                "text_value": concl_text,
                "task_language": language
            }
        ]
        return concl

    def _postprocess_concl(self, concl: list) -> List[dict]:
        """ Postprocess conclusions: join any consecutive conclusions,
        otherwise remove unnecessary info.
        """
        if len(concl) > 0:
            if len(concl) > 1:
                if len(set([item["task_language"] for item in concl]))==1:
                    concl_nums = [int(item["sequence_nr"]) for item in concl]
                    if self._is_consecutive(concl_nums):
                        concl = self._join_conclusions(concl)
            else:
                concl = [
                    {
                        "text_value": item["text_value"],
                        "task_language": item["task_language"]
                    }
                    for item in concl
                ]
        return concl


    def _look_through_section_titles(self,
            digitized_texts: List[dict], search_type: str
    ) -> List[dict]:
        """ Looking through section titles to get the included text.
        """
        LOGGER.debug(f"Going through section titles with search_type='{search_type}'.")
        found_items = []

        has_title = [
            txt for txt in digitized_texts
            if "section_title" in txt
        ]

        has_title_not_none = [
            txt for txt in has_title
            if txt["section_title"] != None
        ]

        for section in has_title_not_none:
            if section["sequence_nr"] not in self.used_sections:
                if len(section["text"])>20:
                    section_title = section["section_title"].lower().strip().strip(".,")
                    if section_title in self.search_types_keywords[search_type]:
                        if search_type == TextPartLabel.TABLE_OF_CONTENTS:
                            found_items.append(
                                self._remove_double_toc(section["text"])
                            )
                            self.used_sections.append(section["sequence_nr"])
                        elif search_type == TextPartLabel.CONCLUSION:
                            found_items.append(
                                {
                                    "sequence_nr": section["sequence_nr"],
                                    "text_value":section["text"],
                                    "task_language":section["language"]
                                }
                            )
                            self.used_sections.append(section["sequence_nr"])
                        else:
                            found_items.append(
                                {
                                    "text_value": section["text"],
                                    "task_language":section["language"]
                                }
                            )
                            self.used_sections.append(section["sequence_nr"])
        if search_type == TextPartLabel.CONCLUSION:
            found_items = self._postprocess_concl(found_items)
        return found_items

    def _look_through_texts(self,
            digitized_texts: List[dict], search_type: str
    ) -> List[dict]:
        """ Additionally looking through the start of texts
        (might have false positives).
        """
        LOGGER.debug(f"Going through digitized_texts with search_type='{search_type}'.")
        found_items = []
        for section in digitized_texts:
            # We don't look though found sections again
            if section["sequence_nr"] not in self.used_sections:
                if len(section["text"])>20:
                    # We check to see if keywords are in the first
                    # 20 characters of the text
                    text_start = section["text"][0:20]
                    if search_type == TextPartLabel.CONCLUSION:
                        st_matches = [
                            kw for kw in self.conclusion_upper_keywords
                            if kw in text_start
                        ]
                        if st_matches != []:
                            found_items.append(
                                {
                                    "sequence_nr": section["sequence_nr"],
                                    "text_value": section["text"],
                                    "task_language": section["language"]
                                }
                            )
                            self.used_sections.append(section["sequence_nr"])
                    else:
                        st_matches = [
                            kw for kw in self.search_types_keywords[search_type]
                            if kw in text_start
                        ]
                        if st_matches:
                            if search_type == TextPartLabel.TABLE_OF_CONTENTS:
                                found_items.append(
                                    self._remove_double_toc(section["text"])
                                )
                                self.used_sections.append(section["sequence_nr"])
                            else:
                                found_items.append(
                                    {
                                        "text_value": section["text"],
                                        "task_language":section["language"]
                                    }
                                )
                                self.used_sections.append(section["sequence_nr"])
        if search_type == TextPartLabel.CONCLUSION:
            found_items = self._postprocess_concl(found_items)
        return found_items

    def _results_correct_fields(self, results: dict) -> dict:
        LOGGER.debug(f"Updating field names and format.")
        #final_results = {}
        text_parts = []

        try:
            _text_parts = {
                TextPartLabel.TABLE_OF_CONTENTS: ["\n".join(results["table_of_contents"])],
                TextPartLabel.CONCLUSION: results["conclusions"],
                TextPartLabel.ABSTRACT: results["abstracts"]
            }
        except:
            _text_parts = {
                TextPartLabel.TABLE_OF_CONTENTS: [
                    "\n".join([
                        v.get("title", [])[0]
                        for v in results["table_of_contents"]
                        if v.get("title", [])
                    ])
                ],
                TextPartLabel.CONCLUSION: results["conclusions"],
                TextPartLabel.ABSTRACT: results["abstracts"]
            }
        #print(_text_parts)
        for label, values in _text_parts.items():

            for value in values:
                if isinstance(value, dict):
                    value =  value.get("text_value", value)
                if value:
                    new_text_part = TextPart(
                        text_type=label,
                        text_value=value
                    ).to_dict()
                    text_parts.append(new_text_part)

        return text_parts



import regex as re
from copy import deepcopy
from typing import List, Dict, NoReturn, Any, Tuple
from collections import defaultdict
from rara_meta_extractor.tools.utils import detect_language
from rara_meta_extractor.config import (
    LOGGER, META_FIELDS, AUTHOR_FIELDS,
    META_YEAR_FIELDS, AUTHOR_ROLES_DICT
)
from rara_meta_extractor.constants.data_classes import (
    AuthorField, MetaField, TextBlock, DataRestrictions, MetaField, AuthorType,
    AuthorNameOrder, TitleType, TitleTypeSimple, AUTHOR_TYPES_MAP, TITLE_TYPES_MAP,
    META_FIELDS_WITH_MULTIPLE_VALUES
)

difference = lambda x, y: set(x) - set(y)

class MetaValidator:
    def __init__(self, meta_fields: List[str] = META_FIELDS,
            author_fields: List[str] = AUTHOR_FIELDS
    ) -> NoReturn:
        self.meta_fields: List[str] = meta_fields
        self.author_fields: List[str] = author_fields

    def _filter_by_length(self,
        key: str, values: List[Any], length: int, values_to_str: bool = False
    ) -> List[Any]:
        """ Filters out values not complying to length requirements.
        """
        original_values = deepcopy(values)
        filtered_values = []
        for value in values:
            if len(str(value)) == length:
                if values_to_str:
                    filtered_values.append(str(value))
                else:
                    filtered_values.append(value)
        diff = difference(original_values, filtered_values)
        if diff:
            LOGGER.debug(
                f"Removed the following values for key '{key}' not complying " \
                f" to length requirement ({length} characters): {list(diff)}"
            )
        return filtered_values

    def _remove_empty(self, key: str, values: List[Any]) -> List[Any]:
        """ Remove empty strings and "None"-s.
        """
        original_values = deepcopy(values)
        values = [
            v for v in values
            if str(v).strip() and str(v).strip() != "None"
        ]
        n_removed = len(original_values) - len(values)
        if n_removed > 0:
            LOGGER.debug(f"Removed {n_removed} empty values for key '{key}'.")
        return values

    def _get_validated_values(self, key: str, values: list,
            check_dates: bool = True
    ) -> List[Any]:
        values = self._remove_empty(key, values)

        if key == MetaField.ISBN:
            values = self._filter_by_length(
                key=key,
                values=values,
                length=DataRestrictions.ISBN_LENGTH,
                values_to_str=True
            )
        elif key == MetaField.ISSN:
            values = self._filter_by_length(
                key=key,
                values=values,
                length=DataRestrictions.ISSN_LENGTH,
                values_to_str=True
            )
        elif check_dates and key.strip() in META_YEAR_FIELDS:
            values = self._filter_by_length(
                key=key,
                values=values,
                length=DataRestrictions.YEAR_LENGTH,
                values_to_str=False
            )
        return values

    def _is_valid_key(self, key: str) -> bool:
        """ Checks, if the key is a valid key present in either
        META_FIELDS or AUTHOR_FIELDS list.
        """
        split_key = " ".join(key.split("_"))
        if key in META_FIELDS or split_key in META_FIELDS:
            return True
        if key in AUTHOR_FIELDS or split_key in AUTHOR_FIELDS:
            return True
        return False


class MetaFormatter:
    """ Formats meta
    """
    def __init__(self,
        meta_fields: List[str] = META_FIELDS,
        author_fields: List[str] = AUTHOR_FIELDS
    ) -> NoReturn:
        self.meta_validator = MetaValidator(
            meta_fields=META_FIELDS,
            author_fields=AUTHOR_FIELDS
        )

    def _filter_by_ratio(self, values: List[Tuple[Any, int]],
            n_trials: int, min_ratio: float
    ) -> List[Any]:
        """ Keeps only values that have been predicted
        in `min_ratio` trials.
        """
        filtered_values = [
            v[0] for v in values
            if float(v[1]/n_trials) >= min_ratio
        ]
        return filtered_values

    def _merge_meta(self, meta_batches: List[dict], min_ratio: float = 0.5) -> dict:
        """ Merges meta into a single dict.
        """
        LOGGER.debug("Merging and formatting metadata.")

        formatted_meta = [
            self._format_meta(meta_dict)
            for meta_dict in meta_batches
        ]
        meta = {}
        frequencies = defaultdict(lambda: defaultdict(int))
        n_trials = len(meta_batches)

        for meta_dict in formatted_meta:
            for key, values in meta_dict.items():
                for value in values:
                    frequencies[key][value]+=1

        for key, value_dict in frequencies.items():
            value_list = sorted(
                list(value_dict.items()),
                key=lambda x: x[1],
                reverse=True
            )
            meta[key] = self._filter_by_ratio(
                values=value_list,
                n_trials=n_trials,
                min_ratio=min_ratio
            )
        return meta

    def _parse_authors(self, authors_list: List[str]) -> List[str]:
        """ Parse weirdly extracted authors like:
        "Reelika RätsepMari-Liis TammikElmar Zimmer"
        """
        LOGGER.debug("Parsing authors.")
        new_authors = []
        for author in authors_list:
            parsed = re.split("(?<=[a-züõöäšž])(?=[A-ZÜÕÖÄŠŽ])", author)
            new_authors.extend(parsed)
        return new_authors

    def _parse_dates(self, values: List[str]) -> List[str]:
        parsed = []
        for value in values:
            match = re.search("(?<=\D)\d{4}(?=\D)", value)
            if match:
                parsed.append(match.group())
        return parsed

    def _add_missing_keys(self, meta: dict) -> dict:
        """ Adds missing meta keys.
        """
        meta_copy = deepcopy(meta)
        missing_keys = {}
        for key in META_FIELDS:
            if key not in meta:
                missing_keys[key] = []
        meta_copy.update(missing_keys)
        return meta_copy

    def _remove_empty_fields(self, meta: dict) -> dict:
        """ Removes empty fields.
        """
        new_meta = {}
        for key, values in meta.items():
            if values:
                new_meta[key] = values
        return new_meta

    def _format_meta(self, meta: dict, check_dates: bool = False) -> dict:
        """ Format meta.
        """
        formatted_meta = {}
        for key, values in list(meta.items()):
            key = key.strip()
            if not self.meta_validator._is_valid_key(key):
                LOGGER.error(
                    f"Detected an invalid key '{key}'. " \
                    f"This will NOT be added to the output."
                )
                continue
            values = self.meta_validator._get_validated_values(
                key=key, values=values, check_dates=check_dates
            )
            if values:
                if key in AUTHOR_FIELDS:
                    formatted_meta[key] = self._parse_authors(values)
                elif key in META_YEAR_FIELDS:
                    formatted_meta[key] = self._parse_dates(values)
                else:
                    formatted_meta[key] = values
        return formatted_meta


    def _format_authors(self, meta: dict, simple: bool = False) -> dict:
        """ Convert authors from a flat structure into
        a list of dicts.
        """
        LOGGER.debug("Formatting authors.")
        new_meta = {MetaField.AUTHORS: []}#, "main_author": ""}
        for key, values in list(meta.items()):
            if key in AUTHOR_FIELDS:
                for i, value in enumerate(values):
                    if i == 0 and key == AuthorField.AUTHOR:
                        is_primary = True
                    else:
                        is_primary = False
                    new_author = Author(
                        name=value,
                        role=key,
                        is_primary=is_primary
                    ).to_dict(simple=simple)
                    new_meta["authors"].append(new_author)
            else:
                new_meta[key] = values
        return new_meta


class Author:
    def __init__(self, name: str, role: str, is_primary: bool, author_type: str = AuthorType.UNK,
            unknown_role: str = AuthorField.UNKNOWN, map_role: bool = True,
            author_name_order: int = AuthorNameOrder.FIRST_NAME_FIRST
    ):
        self.name: str = name
        self.is_primary: bool = is_primary
        self.en_role: str = role
        self.et_role: str = AUTHOR_ROLES_DICT.get(self.en_role, unknown_role) if map_role else role
        self.author_type: str = AUTHOR_TYPES_MAP.get(author_type, AuthorType.UNK)
        self.author_name_order: int = author_name_order

    def to_dict(self, simple: bool = False) -> dict:
        if simple:
            author_dict = {
                "name": self.name,
                "role": self.et_role
            }
        else:
            author_dict = {
                "is_primary": self.is_primary,
                "name": self.name,
                "role": self.et_role,
                "type": self.author_type,
                "name_order": self.author_name_order
            }
        return author_dict

class Title:
    def __init__(self, titles: List[str], lang: str = "",
        part_number: str = "", title_type: str = ""
    ):
        self.titles = titles
        self.lang = lang
        self.part_number = part_number
        self.title_type = title_type

    def _find_skip(self, title: str) -> int:
        # TODO!
        return 0

    def _get_title_type(self, i: int, simple: bool) -> str:
        if self.title_type:
            return self.title_type
        if simple:
            title_type_class = TitleTypeSimple
        else:
            title_type_class = TitleType
        if i == 0:
            title_type = title_type_class.TITLE
        if i >= 1:
            title_type = title_type_class.ADDITIONAL_TITLE
        return title_type

    def to_dicts(self, simple: bool = False) -> List[dict]:
        formatted = []
        for i, title in enumerate(self.titles):
            if i > 1 and not simple:
                break

            title_type = self._get_title_type(i=i, simple=simple)
            title_type_int = TITLE_TYPES_MAP.get(title_type, None)
            skip = self._find_skip(title)
            title_language = detect_language(title)

            if not simple:
                title_dict = {
                    "title": title,
                    "title_language": title_language,
                    "part_number": self.part_number,
                    "part_title": "",
                    "version": "",
                    "author_from_title": "",
                    "skip": skip,
                    "title_type": title_type,
                    "title_type_int": title_type_int
                }
            else:
                title_dict = {
                    "title": title,
                    "title_type": title_type
                }
            formatted.append(title_dict)
        return formatted

class TextPart:
    def __init__(self, text_type: str, text_value: str, language: str = ""):
        self.text_type = text_type
        self.text_value = text_value
        self.language = language if language else detect_language(text_value)

    def to_dict(self) -> dict:
        text_part_dict = {
            "text_type": self.text_type,
            "text_value": self.text_value,
            "language": self.language
        }
        return text_part_dict

class Meta:
    def __init__(self, meta_batches: List[dict], text_parts: List[dict], min_ratio: float = 0.8,
        add_missing_keys: bool = False, simple: bool = False
    ) -> NoReturn:
        self.meta_batches: List[dict] = meta_batches
        self.text_parts: List[dict] = text_parts
        self.min_ratio: float = min_ratio
        self.add_missing_keys: bool = add_missing_keys
        self.simple: bool = simple

        self.meta_formatter: MetaFormatter = MetaFormatter(
            meta_fields = META_FIELDS,
            author_fields = AUTHOR_FIELDS
        )

        self.__merged_meta: dict = {}
        self.__meta_with_reformatted_authors: dict = {}
        self.__meta_with_all_keys: dict = {}
        self.__meta_without_empty_fields: dict = {}

    @property
    def merged_meta(self) -> dict:
        if not self.__merged_meta:
            self.__merged_meta = self.meta_formatter._merge_meta(
                meta_batches=self.meta_batches,
                min_ratio=self.min_ratio
            )
        return self.__merged_meta

    @property
    def meta_with_reformatted_authors(self):
        if not self.__meta_with_reformatted_authors:
            self.__meta_with_reformatted_authors = self.meta_formatter._format_authors(
                meta=self.merged_meta, simple=self.simple
            )
        return self.__meta_with_reformatted_authors

    @property
    def meta_with_all_keys(self):
        if not self.__meta_with_all_keys:
            self.__meta_with_all_keys = self.meta_formatter._add_missing_keys(
                meta=self.meta_without_empty_fields
            )
        return self.__meta_with_all_keys

    @property
    def meta_without_empty_fields(self):
        if not self.__meta_without_empty_fields:
            self.__meta_without_empty_fields = self.meta_formatter._remove_empty_fields(
                meta=self.meta_with_reformatted_authors
            )
        return self.__meta_without_empty_fields

    def _concat(self, meta: dict) -> dict:
        concatted = {}
        for key, value in meta.items():
           key_tokens = key.split()
           new_key = "_".join(key_tokens)
           concatted[new_key] = value
        return concatted

    @staticmethod
    def update_field_types(meta_dict: dict, empty_value: Any = "",
            custom_keys_to_ignore: List[str] = []
    ) -> dict:
        keys_to_ignore = [MetaField.TITLES, MetaField.AUTHORS, MetaField.TEXT_PARTS]
        keys_to_ignore.extend(META_FIELDS_WITH_MULTIPLE_VALUES)
        keys_to_ignore.extend(custom_keys_to_ignore)
        for key, values in list(meta_dict.items()):
            if key not in keys_to_ignore:
                if not isinstance(values, list):
                    continue
                if values:
                    meta_dict[key] = values[0]
                else:
                    meta_dict[key] = empty_value
        return meta_dict

    def to_dict(self):
        if self.add_missing_keys:
            meta_dict = self._concat(self.meta_with_all_keys)
        else:
            meta_dict = self._concat(self.meta_without_empty_fields)
        if self.text_parts or self.add_missing_keys:
            meta_dict[MetaField.TEXT_PARTS] = self.text_parts
        if MetaField.TITLES in meta_dict:
            titles = meta_dict.pop(MetaField.TITLES)
            formatted_titles = Title(titles=titles).to_dicts(simple=self.simple)
            meta_dict[MetaField.TITLES] = formatted_titles

        meta_dict = Meta.update_field_types(meta_dict)
        # Hotfix
        keys_to_remove = ["country_from_008", "unknown"]
        new_dict = {}
        for key, values in meta_dict.items():
            if key not in keys_to_remove:
                new_dict[key] = values
        meta_dict = new_dict

        return meta_dict



import regex as re
from typing import List
from rara_meta_extractor.config import LOGGER
from rara_meta_extractor.constants.data_classes import TextPartLabel
from rara_meta_extractor.constants.patterns import (
    TABLE_OF_CONTENTS_PATTERNS, ABSTRACT_PATTERNS,
    CONCLUSION_PATTERNS, ORDINAL_PATTERNS
)


class RegexTextPartClassifier:
    def __init__(self):
        # Table of contents item variants
        self.regex_pattern_toc_items1: str = r"\b[\p{Ll}\p{Lu}]+\s+(\. ?){4,}\s+[0-9]+"
        self.regex_pattern_toc_items2: str = r"\b[\p{Ll}\p{Lu}]+\s+[0-9]+"
        self.regex_pattern_toc_items3: str = r"\b[\p{Ll}\p{Lu}0-9]+\n([0-9]+\n){3,}"

        self.ordinal_pattern: str = rf"|".join(ORDINAL_PATTERNS)

        self.__toc_variants: str | re.Pattern = ""
        self.__abstract_variants: str | re.Pattern = ""
        self.__conclusion_variants: str | re.Pattern = ""

    @property
    def toc_variants(self) -> re.Pattern:
        """ Constructs and compiles pattern for
        detecting tables of content.

        Returns
        ---------
        re.Pattern
            Compiled pattern for detecting tables of content.
        """
        if not self.__toc_variants:
            _toc_variants = self._create_variants(
                list_of_items=TABLE_OF_CONTENTS_PATTERNS,
                items_type=TextPartLabel.TABLE_OF_CONTENTS
            )
            self.__toc_variants = re.compile(_toc_variants)
        return self.__toc_variants

    @property
    def abstract_variants(self) -> re.Pattern:
        """ Constructs and compiles pattern for
        detecting abstracts.

        Returns
        ---------
        re.Pattern
            Compiled pattern for detecting abstracts.
        """
        if not self.__abstract_variants:
            _abstract_variants = self._create_variants(
                list_of_items=ABSTRACT_PATTERNS,
                items_type=TextPartLabel.ABSTRACT
            )
            self.__abstract_variants = re.compile(_abstract_variants)
        return self.__abstract_variants

    @property
    def conclusion_variants(self) -> re.Pattern:
        """ Constructs and compiles pattern for
        detecting conclusions.

        Returns
        ---------
        re.Pattern
            Compiled pattern for detecting conclusions.
        """
        if not self.__conclusion_variants:
            _conclusion_variants = self._create_variants(
                list_of_items=CONCLUSION_PATTERNS,
                items_type=TextPartLabel.CONCLUSION
            )
            self.__conclusin_variants = re.compile(_conclusion_variants)
        return self.__conclusion_variants

    def _make_uppercase_variant(self, item: str) -> str:
        """ Makes uppercase variant of an item.

        Parameters
        -----------
        item: str
            Item to uppercase.

        Returns
        ----------
        str:
            An uppercased item.
        """
        return item.upper()


    def _add_boundaries_front(self, item: str) -> str:
        """ Adds word boundary to the front.

        Parameters
        -----------
        item: str
            Item to add boundary to.

        Returns
        ----------
        str:
            Item with a front boundary.
        """
        return r"\b"+item


    def _add_boundaries_front_newline(self, item: str) -> str:
        """ Adds word boundary to the front with
        newline (in the case of only conclusions).

        Parameters
        -----------
        item: str
            Item to add boundary to.

        Returns
        ----------
        str:
            Item with a front newline boundary.
        """
        return r"((?<=\n)|(?<=^))(( ?[0-9]+\.? ?)+)?(( ?[VIX]+\.? ?)+)?([\t\r ]+)?" + item


    def _add_boundaries_back(self, item: str) -> str:
        """ Adds word boundary to the front.

        Parameters
        -----------
        item: str
            Item to add boundary to.

        Returns
        ----------
        str:
            Item with an end boundary.
        """
        return item+r"\b"


    def _attach_variants(self, list_of_items: List[str]) -> str:
        """ Attaches variants for regular expression pattern.

        Parameters
        -----------
        list_of_items: List[str]
            Items to concatinate.

        Returns
        ----------
        str:
            Concatinated items.
        """
        return "|".join(list_of_items)


    def _create_variants(self, list_of_items: List[str], items_type: str) -> str:
        """ Creates different variants depending on the item types.
        Items' type is either conclusion or other.

        Parameters
        -----------
        list_of_items: List[str]
            List of patterns / strings.
        items_type: str
            TextPartLabel class param value.
        """
        LOGGER.debug(f"Generating variations for items with type '{items_type}'.")
        uppercase_variants = [
            self._make_uppercase_variant(item)
            for item in list_of_items
        ]

        list_of_items.extend(uppercase_variants)

        if items_type != TextPartLabel.CONCLUSION:
            boundaries_front = [
                self._add_boundaries_front(item)
                 for item in list_of_items
             ]
            boundaries_back = [
                self._add_boundaries_back(item)
                for item in list_of_items
            ]
            with_boundaries = boundaries_front + boundaries_back
            variants =  self._attach_variants(with_boundaries)
        else:
            with_boundaries = [
                self._add_boundaries_front_newline(item)
                for item in list_of_items
            ]
            variants =  self._attach_variants(with_boundaries)
        return variants


    def is_table_of_contents(self, text: str) -> bool:
        """ Determine, if text contains table of contents.

        Parameters
        -----------
        text: str
            Input text.

        Returns
        -----------
        bool
            Boolean value indicating, if text contains table of contents.
        """
        if re.search(self.regex_pattern_toc_items1, text):
                return True
        else:
            if re.search(self.toc_variants, text):
                if re.search(self.regex_pattern_toc_items2, text):
                    return True
                elif re.search(self.regex_pattern_toc_items3, text):
                    return True
                return True
            elif re.search(self.regex_pattern_toc_items2, text):
                if re.search(self.toc_variants, text.upper()):
                    return True
            elif re.search(self.regex_pattern_toc_items3, text):
                if re.search(self.toc_variants, text.upper()):
                   return True
        return False

    def is_abstract(self, text: str) -> bool:
        """ Determine, if text contains an abstract.

        Parameters
        -----------
        text: str
            Input text.

        Returns
        -----------
        bool
            Boolean value indicating, if text contains an abstarct.
        """
        if re.search(self.abstract_variants, text):
            return True
        return False

    def is_conclusion(self, text: str) -> bool:
        """ Determine, if text contains a conclusion.

        Parameters
        -----------
        text: str
            Input text.

        Returns
        -----------
        bool
            Boolean value indicating, if text contains a conclusion.
        """
        # first we check if any conclusion variants in the text,
        # since the pattern is very long and complex
        simple_pattern = re.compile(self._attach_variants(CONCLUSION_PATTERNS))
        #concl_titles = re.compile(self.conclusin_variants)
        if re.search(simple_pattern, text):
            if re.search(self.conclusion_variants, text):
                get_specific_part = [
                    line
                    for line in text.split("\n")
                    if re.search(self.conclusion_variants, line)
                ]
                for matched_line in get_specific_part:
                    # Trying to avoid chapter conclusions
                    if not re.search(self.ordinal_pattern, matched_line.lower()):
                        # Checking if there's any text after the conclusion title
                        if matched_line != text.split("\n")[-1]:
                            next_line = text.split("\n")[text.split("\n").index(matched_line)+1]
                            if len(next_line.strip())==0:
                                line_after = text.split("\n")[text.split("\n").index(matched_line)+2]
                                if len(line_after) > 35:
                                    return True
                            elif len(next_line) > 35:
                                return True
                        else:
                            return True
        return False

    def get_label(self, text: str) -> str:
        """
        Determines text part in order of exclusion,
        single doc can't have multiple text parts

        Parameters
        -----------
        text: str
            Input text.

        Returns
        -----------
        str:
            Text part label.
        """
        LOGGER.info(f"Detecting text part for text {text[:20]}...")
        if self.is_table_of_contents(text):
            label = TextPartLabel.TABLE_OF_CONTENTS
        elif self.is_abstract(text):
            label = TextPartLabel.ABSTRACT
        elif self.is_conclusion(text):
            label = TextPartLabel.CONCLUSION
        else:
            label = TextPartLabel.OTHER
        return label



import regex as re
from typing import List, Dict
from collections import defaultdict
from rara_meta_extractor.config import (
    META_EXTRACTOR_CONFIG, TEXT_CLASSIFIER_CONFIG, METADATA_TEXT_BLOCKS, LOGGER,
    AUTHOR_FIELDS, META_FIELDS, TextBlock
)
from rara_meta_extractor.constants.data_classes import (
    ExtractorType, MetaField, TextBlock, AuthorField, TextPartLabel
)
from rara_meta_extractor.tools.meta_formatter import Meta, TextPart
from rara_meta_extractor.tools.utils import detect_language
from rara_meta_extractor.llm_agents import TextClassifierAgent, MetaExtractorAgent
from rara_meta_extractor.epub_meta_extractor import EPUBMetaExtractor
from rara_meta_extractor.mets_alto_meta_extractor import MetsAltoMetaExtractor
from rara_meta_extractor.text_part_classifiers.regex_text_part_classifier import RegexTextPartClassifier
from rara_meta_extractor.text_part_classifiers.epub_text_part_classifier import EPUBTextPartClassifier
from rara_meta_extractor.text_part_classifiers.mets_alto_text_part_classifier import MetsAltoTextPartClassifier


class MetaExtractor:
    def __init__(self,
        meta_extractor_config: dict = META_EXTRACTOR_CONFIG,
        text_classifier_config: dict = TEXT_CLASSIFIER_CONFIG
    ):
        self.meta_extractor_config = meta_extractor_config
        self.text_classifier_config = text_classifier_config

        self.meta_agent = MetaExtractorAgent(self.meta_extractor_config)
        self.text_classifier_agent = TextClassifierAgent(self.text_classifier_config)

        self.text_part_classifier = RegexTextPartClassifier()
        self.epub_text_part_classifier = EPUBTextPartClassifier()
        self.mets_alto_text_part_classifier = MetsAltoTextPartClassifier()

        self.epub_meta_extractor = EPUBMetaExtractor()
        self.mets_alto_meta_extractor = MetsAltoMetaExtractor()



    def classify_text(self, text: str, default: str = TextBlock.METADATA) -> str:
        """ Classifies text into one of the text blocks defined in TextBlock.

        Parameters
        -----------
        text: str
            Text to classify
        default: str
            Default value to return in case of an exception.

        Returns
        -----------
        str:
            Text class.
        """
        try:
            text_type_dict = self.text_classifier_agent.extract(text=text)
            LOGGER.debug(f"Llama text classifier output: {text_type_dict}")
            print(text_type_dict)
            text_class = text_type_dict.get("text_type")[0]
        except Exception as e:
            LOGGER.error(
                f"Detecting text type for text '{text[:50]}' failed with error: {e}. "
                f"Defaulting to '{default}'"
            )
            text_class = default
        return text_class



    def _construct_llama_input(self, texts: List[str], use_llm: bool = False,
            max_length_per_text: int = 1500, n_first_pages: int = 5,
            n_last_pages: int = 0, n_strict_include: int = 3
    ) -> List[str]:
        """ Construct input for Llama.

        Parameters
        -----------
        texts: List[str]
            Raw texts
        use_llm: bool
            If enabled, LLM is used for each n_first_page and n_last_page text
            to detect, if they contain metadata or not. Only texts containing
            metadata will be passed to Llama for metadata extraction.
        max_length_per_text: int
            Used only if `use_llm=False` and is not applied to `n_strict_include`
            first texts. For other texts, this param will be used to decide, if
            the text will be added to Llama input or not (texts longer than the
            threshold set with this param will be ignored).
        n_first_pages: int
            How many first pages to consider as potential inputs?
        n_last_pages: int
            How many last pages to consider as potential inputs?
        n_strict_include: int
            How many first pages to always include?

        Returns
        -----------
        List[str]:
            List of verified texts
        """
        LOGGER.debug("Constructing Llama input!")
        n_pages  = len(texts)

        # If number of pages is smaller than number of
        # first and last pages combined, modify the
        # number of last pages in order to avoid
        # duplicates in final Llama input
        if n_pages < n_first_pages + n_last_pages:
            n_last_pages = max(n_pages - n_first_pages, 0)

        first_pages = texts[:n_first_pages]
        last_pages = texts[-n_last_pages:] if n_last_pages > 0 else []
        if n_strict_include > n_first_pages:
            n_strict_include = n_first_pages
        texts_to_analyze = first_pages + last_pages

        if use_llm:
            verified_texts = []
            for text in texts_to_analyze:
                text_type = self.classify_text(text)
                if text_type in METADATA_TEXT_BLOCKS:
                    verified_texts.append(text)
            if not verified_texts:
                LOGGER.error(
                    f"No verified metadata text block found from texts {texts} with LLM!"
                )
        else:
            verified_texts = texts_to_analyze[:n_strict_include]

            for text in texts_to_analyze[n_strict_include:]:
                if (len(text) < max_length_per_text and not
                    re.search("sisukord|table of contents", text, re.IGNORECASE)
                ):
                    verified_texts.append(text)
        return verified_texts

    def _get_text_parts(self, texts: List[dict]) -> List[dict]:
        """ Wrapper function for detecting text partsself.

        Parameters
        -----------
        texts: List[dict]
            Raw texts.

        Returns
        -----------
        List[dict]:
            List of dicts where each dict contains
            one text part label, corresponding text block and
            language of the text block. NB! multiple texts might be concatted
            into one, if conclusion spans over multiple pages etc.

        """
        text_parts = []
        _text_parts = defaultdict(lambda: defaultdict(list))

        for text in texts:
            raw_text = text.get("text", "")
            lang = text.get("language", "")
            label = self.text_part_classifier.get_label(raw_text)
            if label != TextPartLabel.OTHER:
                _text_parts[label]["texts"].append(raw_text)
                _text_parts[label]["langs"].append(lang)

        for label, values in _text_parts.items():
            texts = values.get("texts")
            langs = values.get("langs")
            lang_counts = defaultdict(int)
            for lang in langs:
                lang_counts[lang]+=1
            most_frequent_lang = sorted(
                list(lang_counts.items()),
                key=lambda x: x[1],
                reverse=True
            )[0][0]
            if not most_frequent_lang:
                most_frequent_lang = detect_language(text_part)
            text_part = "\n".join(texts)
            new_text_part = TextPart(
                text_type=label,
                text_value=text_part,
                language=most_frequent_lang
            ).to_dict()

            text_parts.append(new_text_part)
        return text_parts

    def extract_from_digitizer_output(self, digitizer_output: dict, **kwargs) -> dict:
        """ Extracts metadata from Digitizer output.
        """
        LOGGER.info("Extracting metadata from Digitizer output.")

        texts = digitizer_output.get("texts", [])
        doc_meta = digitizer_output.get("doc_meta", {})

        mets_alto_metadata = doc_meta.get("mets_alto_metadata", [])
        epub_metadata = doc_meta.get("epub_metadata", {})

        result = self.extract(
            texts=texts,
            epub_metadata=epub_metadata,
            mets_alto_metadata=mets_alto_metadata,
            **kwargs
        )
        return result

    def extract_simple(self, text: str, lang: str,
            simple: bool = True, **kwargs
    ) -> dict:
        """ For simple usage. Passes plaintext to Llama extractor.
        """
        texts = [{"text": text, "lang": lang}]
        result = self.extract(texts=texts, simple=simple, **kwargs)
        return result

    def extract(self,
            texts: List[dict],
            epub_metadata: dict = {},
            mets_alto_metadata: List[str] = [],
            verify_texts_with_llm: bool = False,
            n_trials: int = 1,
            merge_texts: bool = True,
            min_ratio: float = 0.8,
            add_missing_keys: bool = False,
            detect_text_parts: bool = True,
            max_length_per_text: int = 1500,
            n_first_pages: int = 5,
            n_last_pages: int = 0,
            n_strict_include: int = 3,
            simple: bool = False
    ) -> dict:
        """ Extracts relevant metadata from a batch of texts

        Parameters
        -----------
        texts: List[dict]
            List of texts from where to extract meta information. For EPUB and METS/ALTO,
            expects content of `texts` from digitizer output. Otherwise, must minimally
            contain keys `text` and `lang`.
        epub_metadata: dict
            Expects the content of `doc_meta.epub_metadata` from digitizer output.
        mets_alto_meta: List[str]
            Expects the content of `doc_meta.mets_alto_metadata` from digitizer output.
        verify_texts_with_llm: bool
            If enabled, each text is passed to text classifier agent first
            and only texts classified as metadata blocks are passed to
            meta extractor(s).
        n_trials: int
            Indicates how many trials to run for predicting metadata with LlamaExtractor
            for the same text. NB! Setting it higher than 1 has purpose only if temperature
            > 0.
        merge_texts: bool
            If enabled, texts are merged into a single text block before passing
            it to LlamaExtractor. Otherwise texts are passed one by one to LlamaExtractor
            and results are merged afterwards.
        min_ratio: float
            Relevant only if n_trials > 1. Indicates the ratio of times a meta value
            has to be predicted during trials. E.g. if min_ratio = 0.7 and a value is predicted
            2 out of 3 trials, it will not be returned as 2/3 = 0.66 < 0.7.
        add_missing_keys: bool
            If enabled, all possible meta keys are added to the output, even if
            the content has not been extracted.
        detect_text_parts: bool
            If enabled, runs text part detection for detecting conclusions, abstracts etc.
        max_length_per_text: int
            If verify_texts_with_llm is set to False, this param is used for dummy metadata detection -
            if a text is longer than the threshold set with this param, it will not be included
            into Llama input.
        n_first_pages: int
            How many first pages to consider for possible Llama input? NB! Not all of them are
            actually added to the input as the pages are passed through prefiltering.
        n_last_pages: int
            How many last pages to consider for possible Llama input? NB! Not all of them are
            actually added to the input as the pages are passed through prefiltering.
        n_strict_include: int
            Number of pages (out of n_first_pages + n_list_pages set) to pass to Llama
            without additional prefiltering.
        simple: bool
            If enabled, the outputs of titles and authors are simplified (some fields
            necessary mostly for constructing final MARC files are removed).

        """
        meta_batches = []
        text_parts = []
        raw_meta = {}

        # TODO: param for applying LLM if mets-alto / epub meta is present?
        # or make it dynamic based on what fields are detected?
        if epub_metadata:
            extractor_type = ExtractorType.EPUB
            LOGGER.info(f"Using {extractor_type} for metadata extraction.")
            meta: dict = self.epub_meta_extractor.extract_meta(
                epub_metadata=epub_metadata, simple=simple
            )

        elif mets_alto_metadata:
            extractor_type = ExtractorType.METS_ALTO
            LOGGER.info(f"Using {extractor_type} for metadata extraction.")
            meta: dict = self.mets_alto_meta_extractor.extract_meta(
                mets_alto_metadata=mets_alto_metadata,
                digitized_texts=texts,
                simple=simple
            )

        else:
            extractor_type = ExtractorType.LLAMA
            LOGGER.info(f"Using {extractor_type} for metadata extraction.")
            raw_texts = [
                doc.get("text") for doc in texts
                if doc.get("text").strip()
            ]

            verified_texts = self._construct_llama_input(
                texts=raw_texts,
                use_llm=verify_texts_with_llm,
                max_length_per_text=max_length_per_text,
                n_first_pages=n_first_pages,
                n_last_pages=n_last_pages,
                n_strict_include=n_strict_include
            )

            if merge_texts:
                text_str = "\n\n".join(verified_texts)
                if text_str:
                    verified_texts = [text_str]
                else:
                    verified_texts = []

                LOGGER.debug(f"Constructed Llama input of size {len(text_str)} characters.")

            for text in verified_texts:
                for trial_nr in range(n_trials):
                    try:
                        LOGGER.debug(
                            f"Trial nr {trial_nr}. Extracting information with Llama agent " \
                            f"from text '{text[:20]}...'."
                        )
                        meta_batch = self.meta_agent.extract(text=text)
                        LOGGER.debug(f"Raw LLM output: {meta_batch}")
                        meta_batches.append(meta_batch)
                    except Exception as e:
                        LOGGER.error(
                            f"Extracting meta information from text: {text[:50]} " \
                            f"failed with error: {e}."
                    )
        # Detect and add text parts
        if detect_text_parts:

            if extractor_type == ExtractorType.EPUB:
                LOGGER.info(f"Detecting text parts with EPUB text part classifier.")
                text_parts = self.epub_text_part_classifier.get_parts_of_text(
                    digitized_texts=texts,
                    epub_meta=meta
                )
                meta[MetaField.TEXT_PARTS] = text_parts

            elif extractor_type == ExtractorType.METS_ALTO:
                LOGGER.info(f"Detecting text parts with METS/ALTO text part classifier.")
                text_parts = self.mets_alto_text_part_classifier.get_parts_of_text(
                    digitized_texts=texts,
                    mets_alto_meta=raw_meta
                )
                meta[MetaField.TEXT_PARTS] = text_parts
            else:
                LOGGER.info(f"Detecting text parts with Regex text part classifier.")
                text_parts = self._get_text_parts(texts)


        if extractor_type == ExtractorType.LLAMA:

            meta = Meta(
                meta_batches=meta_batches,
                text_parts=text_parts,
                min_ratio=min_ratio,
                add_missing_keys=add_missing_keys,
                simple=simple
            ).to_dict()

        result = {
            "meta": meta,
            "extractor": extractor_type
        }
        return result



from dataclasses import dataclass

@dataclass(frozen=True)
class TextBlock:
    REGULAR: str = "regular text block"
    METADATA: str = "metadata text block"
    TITLE_PAGE: str = "title page"
    ABSTRACT: str = "abstract text block"
    SUMMARY: str = "summary text block"


@dataclass(frozen=True)
class IssueType:
    NEWSPAPER: str = "Ajaleht"
    JOURNAL: str = "Ajakiri"
    PERIODICAL: str = "Jätkväljaanne"
    BOOK: str = "Raamat"

@dataclass(frozen=True)
class ExtractorType:
    LLAMA: str = "Llama-Extractor"
    EPUB: str = "EPUB-Extractor"
    METS_ALTO: str = "METS/ALTO-Extractor"

@dataclass(frozen=True)
class TextPartLabel:
    TABLE_OF_CONTENTS: str = "Sisukord"
    ABSTRACT: str = "Abstrakt"
    CONCLUSION: str = "Kokkuvõte"
    OTHER: str = "Muu"

@dataclass(frozen=True)
class AuthorField:
    AUTHOR: str = "author"
    UNKNOWN: str = "Teadmata"

@dataclass(frozen=True)
class AuthorType:
    PER: str = "PER"
    ORG: str = "ORG"
    UNK: str = ""

@dataclass(frozen=True)
class TitleType:
    AUTHOR_WITHOUT_TITLE: str = "Pealkirjata autor"
    NORMALIZED_TITLE: str = "Normitud eelispealkiri"
    TITLE: str = "Väljaandes esitatud kujul põhipealkiri"
    ADDITIONAL_TITLE: str = "Rööppealkiri"
    METS_TITLE: str = "Artikli või peatüki pealkiri"

@dataclass(frozen=True)
class TitleTypeSimple:
    TITLE: str = "main_title"
    ADDITIONAL_TITLE: str = "additional_title_part"
    METS_TITLE: str = "article_or_chapter_title"

@dataclass(frozen=True)
class AuthorNameOrder:
    FIRST_NAME_FIRST: int = 0
    LAST_NAME_FIRST: int = 1
    FAMILY_NAME_FIRST: int = 3

@dataclass(frozen=True)
class DataRestrictions:
    ISBN_LENGTH: int = 13
    ISSN_LENGTH: int = 8
    YEAR_LENGTH: int = 4

@dataclass(frozen=True)
class MetaYearField:
    DISTRIBUTION_DATE: str = "distribution date"
    PUBLICATION_DATE: str = "publication date"
    MANUFACTURE_DATE: str = "manufacture date"


@dataclass(frozen=True)
class MetaField:
    TITLES: str = "titles"
    SUBTITLE: str = "subtitle"
    ISBN: str = "isbn"
    ISSN: str = "issn"
    EDITION_INFO: str = "edition info/number"
    DISTRIBUTION_PLACE: str = "distribution_place"
    DISTRIBUTER_NAME: str = "distributer name"
    DISTRIBUTION_DATE: str = "distribution date"
    MANUFACTURE_DATE: str = "manufacture date"
    MANUFACTURER: str = "manufacturer"
    MANUFACTURE_PLACE: str = "manufacture place"
    PUBLICATION_DATE: str = "publication date"
    PUBLISHER: str = "publisher"
    PUBLICATION_PLACE: str = "publication place"
    SUMMARY_LANG: str = "summary or abstract language"
    TOC_LANG: str = "table of contents language"
    ORIGINAL_LANG: str = "language original"
    COUNTRY_FROM_008: str = "country from 008"
    TEXT_TYPE: str = "text_type"
    AUTHORS: str = "authors"
    TEXT_PARTS: str = "text_parts"
    ISSUE_TYPE: str = "issue_type"
    UDK: str = "udk"
    UDC: str = "udc"
    UNK: str = "unknown"

TITLE_TYPES_MAP = {
    TitleType.AUTHOR_WITHOUT_TITLE: 130,
    TitleType.NORMALIZED_TITLE: 240,
    TitleType.TITLE: 245,
    TitleType.ADDITIONAL_TITLE: 246
}


ISSUE_STYLE_MAP = {
    "newspaper": IssueType.NEWSPAPER,
    "ajalehed": IssueType.NEWSPAPER,
    "journal": IssueType.JOURNAL,
    "ajakirjad": IssueType.JOURNAL,
    "periodical": IssueType.PERIODICAL,
    "jätkväljanded": IssueType.PERIODICAL,
    "books": IssueType.BOOK,
    "raamatud": IssueType.BOOK
}
AUTHOR_TYPES_MAP = {"personal": AuthorType.PER}

TEXT_BLOCKS = [v for v in vars(TextBlock()).values()]
META_YEAR_FIELDS = [v for v in vars(MetaYearField()).values()]

META_FIELDS_WITH_MULTIPLE_VALUES = [MetaField.ISBN, MetaField.ISSN, MetaField.UDK, MetaField.UDC]
